defmodule Indian do
  @moduledoc """

  Very simple and basic HTTP/1.1 server, only for illustrative
  purposes.

  Note that we rely on the [:gen_tcp
  library](http://erlang.org/doc/man/gen_tcp.html). `gen_tcp` calls are
  prefixed by a colon (Elixir convention for Erlang calls).

  Written by [Stéphane Bortzmeyer](mailto:stephane@bortzmeyer.org).
  """

  # Logging library is standard <https://hexdocs.pm/logger/Logger.html>
  require Logger
  
  def accept(address, port, loglevel \\ :info) do
    Logger.configure([{:utc_log, true}, {:level, loglevel}]) # Funny, invalid level values are apparently interpreted as "all"
    Logger.configure_backend(:console, [{:format, "$time Indian [$level] $levelpad$message\n"}])
    # On a typical OS, listening with "inet6" should enable IPv4 as
    # well (but it may depend on some settings, such as sysctl on
    # Linux). Also, it forces the user to express IPv4 addresses on
    # the command line option '--bind' as '::ffff:192.0.2.1'.
    # Singleton options like "inet6" must be at the beginning.    
    case :gen_tcp.listen(port,
	  [:inet6, mode: :binary, packet: :raw, active: false, reuseaddr: true, ip: address]) do
      {:ok, socket} ->
	Logger.info("Accepting connections on port #{port}") 
	loop_acceptor(socket)
      {:error, :eacces} ->
	Logger.error("Port #{port} not usable (privilege necessary?)")
	exit(:eacces)
      {:error, e} ->
	Logger.error("Error #{e} in listen")
	exit(:shutdown) 
    end
  end

  def loop_acceptor(socket) do
    case :gen_tcp.accept(socket) do
      {:ok, client} ->     Logger.debug("New connection from #{inspect socket}")
        spawn_monitor(Indian, :serve, [client]) # We don't care what happens to the child process
      {:error, e} -> Logger.error("Error #{e} in accept")
        Logger.error("Error #{e} in accept")
        # We continue, it can be temporary (EMFILE, "too many open files", for instance)
    end
    loop_acceptor(socket) 
  end

  def serve(socket) do
    socket
    |> read_request()
    |> parse_request()
    |> build_response()
    |> write_response()
    serve(socket)
  end

  def read_request(socket) do
    context = %{"socket" => socket}
    # We may need several recv to get all the data. We choosed the
    # lazy way, just one recv but this means the server will fail with
    # clients that do not send everything in one operation, such as
    # telnet. The proper way to read would be to merge reading and
    # parsing (using Content-Length, checking for an empty line, etc)
    # but it is more complicated.
      case :gen_tcp.recv(socket, 0) do 
      {:ok, data} ->
	case :inet.peername(socket) do
	  {:ok,  {ip_address, _port_number}} ->
	    addr = to_string(:inet_parse.ntoa(ip_address))
	    Logger.debug("Serving #{addr}, request is #{data}")
	    {:ok, Map.put(context, "address", addr), data}
	  {:error, e} ->
	    Logger.error("Error #{e} in peername")
	    exit(:shutdown)
	end
      {:error, :closed} -> 
	exit(:shutdown)
      {:error, e} -> 
	Logger.error("Error #{e} in recv")
	exit(:shutdown)
    end
    # gen_tcp documentation
    # <http://erlang.org/doc/man/gen_tcp.html#recv-2> tells us that
    # recv can return :ok and the data received, or :error and a
    # reason for the network problem.
  end

  def parse_request({:ok, context, request}) do
    case Indian.Parser.request(request) do
      {:ok, [method, path, version, headers, _body], _, _, _, _} ->
	case String.split(version, ".") do
	  ["1", _] ->
	    h = headers |> Map.new(fn [k, v] -> {k, v} end)
	    {:ok, Map.put(context, "version", version), method, path, h}
	  _ -> {:error, "Invalid HTTP version (we handle version 1 only)"}
	end
      {:error, reason, _, _, _, _} -> {:error, context, reason}
      _ -> {:error, "Invalid request (not enough EOLs?)"}
    end
  end

  # Network problem, no way to return anything to the client
  def parse_request({:error, :closed}) do
    exit(:closed) # Exits the process https://hexdocs.pm/elixir/Kernel.html#exit/1
  end

  def parse_request({:error, msg}) do
    {:error, msg}
  end

  def build_response({:ok, context, method, path, headers}) do
    context = Map.put(context, "method", method)
    context = Map.put(context, "path", path)
    ua = case headers["User-Agent"] do
	   nil -> "Unknown user agent"
	   _ -> headers["User-Agent"]
	 end
    context = Map.put(context, "user-agent", ua)
    connheader = if headers["Connection"] == nil do
      nil
    else
      String.downcase(headers["Connection"])
    end
    # ab -k does not work (it does with other servers and, other test
    # clients with persistent connections work with us), so there is
    # still a bug in persistent connections.
    permconn = case String.split(context["version"], ".") do
	  [_, "0"] ->
	    case connheader do
	      nil -> false # By default, connections are not permanent in HTTP/1.0
	      "keep-alive" ->  true # We are liberal, since it is
				    # mostly a non-standard value. See
				    # RFC 7230, appendix A.1.2
	      _ -> false 
	    end
	  _ -> # Actually, we can have several connection options but we don't handle this case.
	    case connheader do
	      nil -> true # By default, connections are permanent in HTTP 1.1, RFC 9112, section 9.3
	      "close" ->  false
	      "keep-alive" ->  true # Not sure it is defined for HTTP/1.1 but we're liberal here.
	      _ -> false # RFC 9112, section 9.3 is more subtle than that
	    end
    end
    context = Map.put(context, "persistent-connection", permconn)
    # RFC 9112 section 3.2.2 says we MUST accept the absolute form
    # (not just /foobar but also http://www.example.org/foobar). This
    # would imply to know our base URL… We currently have no way to
    # define it. (If so, we could also test the Host: header…)
    result =
      case method do
	"GET" ->
	  case path do
	    "/" ->  {context, 200, "Default page"}
	    "/date" ->
	      {context, 200, "The time is #{Time.to_string(Time.utc_now())}"}
	    "/hello" -> {context, 200, "Hello, #{context["address"]}!"}
	    _ -> {context, 404, "Unknown path #{path}"}
	  end
	"HEAD" -> # May be factor with GET. An array with functions
		  # (to delay evaluation)?
          case path do
	    "/" ->  {context, 200, ""}
	    "/date" -> {context, 200, ""}
	    "/hello" -> {context, 200, ""}
	    _ -> {context, 404, "Unknown path #{path}"}
	  end
	_ -> {context, 405, "Unsupported method #{method}"} 
			   # HTTP methods are not a set fixed for
			   # ever, new methods can be added to the
			   # registry
			   # <https://www.iana.org/assignments/http-methods/http-methods.xml>

	# May be other methods like OPTIONS or POST (this one requires
	# a body)
      end
    context = Map.put(context, "rcode", elem(result, 1))
    # HTTP requires \r\n but \n is the end-of-line on Unix, so we just add \r
    response = """
HTTP/1.1 #{context["rcode"]} \r
Content-Type: text/plain\r
Content-Length: #{byte_size(elem(result, 2))+1}\r
Server: myBeautifulServerWrittenInElixir\r
\r
#{elem(result, 2)}
"""
    {:ok, context, response}
    # You can test that the content-length is correct with curl, that
    # otherwise displays "Excess found in a read: excess = 1, size =
    # 12, maxdownload = 12, bytecount = 0"
  end

  def build_response({:error, context, reason}) do
    context = Map.put(context, "persistent-connection", false)
    context = Map.put(context, "rcode", "400")
    Logger.debug("Wrong request from #{context["address"]} because #{reason}")
    body = "Cannot understand your request because \"#{reason}\""
    response = """
      HTTP/1.1 400 \r
      Content-Type: text/plain\r
      Content-Length: #{byte_size(body)+1}\r
      Server: my beautiful server written in Elixir\r
      \r
      #{body}
      """
    {:ok, context, response}
  end
    
  def write_response({_, context, response}) do
    case :gen_tcp.send(context["socket"], response) do
      :ok -> Logger.info("#{context["address"]} (#{context["user-agent"]}) HTTP/#{context["version"]} (#{context["persistent-connection"]}) #{context["method"]} #{context["path"]} -> #{context["rcode"]}")
      {:error, reason} -> Logger.error("Cannot send response to #{context["address"]}: #{reason}")
      end
    if not context["persistent-connection"] do
      # Since connections are persistent by default in HTTP/1.1, 'echo
      # -n "GET / HTTP/1.1\r\n\r\n" | nc localhost 8080' will keep the
      # connection open, beware.
      :gen_tcp.shutdown(context["socket"], :read_write)
      :gen_tcp.close(context["socket"])
    end
  end

end
